self.__BUILD_MANIFEST=function(s,c,a,e,t,r,n,u,i,b,j,k){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":["static/chunks/pages/index-6b0d9e5031b70c58.js"],"/_error":["static/chunks/pages/_error-1be831200e60c5c0.js"],"/clusters":[s,r,n,c,a,e,t,u,b,"static/chunks/pages/clusters-5549a350f97d7ef3.js"],"/clusters/[cluster]":[s,r,n,c,a,e,t,u,i,b,"static/chunks/pages/clusters/[cluster]-35cbeb5214fd4036.js"],"/clusters/[cluster]/[job]":[s,c,"static/chunks/pages/clusters/[cluster]/[job]-65d04d5d77cbb6b6.js"],"/config":[s,r,c,a,"static/chunks/pages/config-1a1eeb949dab8897.js"],"/infra":[s,c,a,e,t,j,"static/chunks/pages/infra-13b117a831702196.js"],"/infra/[context]":[s,c,a,e,t,j,"static/chunks/pages/infra/[context]-b68ddeed712d45b5.js"],"/jobs":[s,n,c,a,e,t,u,i,"static/chunks/pages/jobs-a76b2700eca236f7.js"],"/jobs/[job]":[s,r,c,e,"static/chunks/pages/jobs/[job]-2d23a9c7571e6320.js"],"/users":[s,c,a,e,t,"static/chunks/pages/users-262aab38b9baaf3a.js"],"/workspace/new":[s,r,n,c,a,e,t,u,i,k,"static/chunks/pages/workspace/new-c7516f2b4c3727c0.js"],"/workspaces":[s,r,n,c,a,e,t,u,i,"static/chunks/pages/workspaces-384ea5fa0cea8f28.js"],"/workspaces/[name]":[s,r,n,c,a,e,t,u,i,k,"static/chunks/pages/workspaces/[name]-7799de9e691e35d8.js"],sortedPages:["/","/_app","/_error","/clusters","/clusters/[cluster]","/clusters/[cluster]/[job]","/config","/infra","/infra/[context]","/jobs","/jobs/[job]","/users","/workspace/new","/workspaces","/workspaces/[name]"]}}("static/chunks/614-635a84e87800f99e.js","static/chunks/470-9e7a479cc8303baa.js","static/chunks/293-351268365226d251.js","static/chunks/969-c7abda31c10440ac.js","static/chunks/856-3a32da4b84176f6d.js","static/chunks/798-c0525dc3f21e488d.js","static/chunks/121-865d2bf8a3b84c6a.js","static/chunks/973-1a09cac61cfcc1e1.js","static/chunks/236-a90f0a9753a10420.js","static/chunks/37-beedd583fea84cc8.js","static/chunks/682-6647f0417d5662f0.js","static/chunks/843-c296541442d4af88.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();