(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[405],{5557:function(n,u,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/",function(){return t(6158)}])},6158:function(n,u,t){"use strict";t.r(u),t.d(u,{default:function(){return r}});var e=t(7294),s=t(1163);function r(){let n=(0,s.useRouter)();return(0,e.useEffect)(()=>{"/"===n.asPath?n.push("/clusters"):n.push(n.asPath)},[n]),null}},1163:function(n,u,t){n.exports=t(6036)}},function(n){n.O(0,[888,774,179],function(){return n(n.s=5557)}),_N_E=n.O()}]);