(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[233],{1803:function(n,t,r){(window.__NEXT_P=window.__NEXT_P||[]).push(["/infra/[context]",function(){return r(1200)}])},1200:function(n,t,r){"use strict";r.r(t),r.d(t,{default:function(){return f}});var u=r(5893);r(7294);var e=r(8682),i=r(9008),c=r.n(i);function f(){return(0,u.jsxs)(u.Fragment,{children:[(0,u.jsx)(c(),{children:(0,u.jsx)("title",{children:"Infra | SkyPilot Dashboard"})}),(0,u.jsx)(e.fx,{})]})}}},function(n){n.O(0,[614,470,293,969,856,682,888,774,179],function(){return n(n.s=1803)}),_N_E=n.O()}]);